from bokeh.palettes import viridis,magma
from bokeh.models import ColumnDataSource, Label, LabelSet,CustomJS
from bokeh.layouts import column,row,gridplot
from bokeh.plotting import figure, show
from bokeh.models import LinearColorMapper, LogTicker,BasicTicker, ColorBar,LogColorMapper,Legend
from bokeh.palettes import magma as colfun1
from bokeh.palettes import Colorblind8
from bokeh.palettes import viridis as colfun2
from bokeh.palettes import gray as colfun3
from bokeh.palettes import Colorblind8
import bokeh.palettes as colpals
import astropy.units as u
import numpy as np

from . import justdoit as pyeddy

def pt(out, with_condensation=True,return_condensation=False, **kwargs):
    """
    Plot PT profiles with condensation curves. 
    Thick lines in this pt plot signify those gases that 
    were turned on in the run, NOT those that were recommended. 
    
    Parameters
    ----------
    out : dict 
        Dictionary output from pyeddy run 
    with_condensation : bool 
        Plots condensation curves of gases. Also plots those that were turned on in the 
        calculation with line_width=5. All others are set to 1. 
    return_condensation : bool 
        If true, it returns list of condenation temps for each gas, pressure grid, 
        and a list of the gas names
    **kwargs : kwargs 
        Kwargs for bokeh.figure() 

    Returns
    -------
    if return_condensation: fig, cond temperature curves, ,cond pressure grid , name of all available gases
    else: fig
    """

    kwargs['plot_height'] = kwargs.get('plot_height',300)
    kwargs['plot_width'] = kwargs.get('plot_width',600)
    kwargs['x_axis_label'] = kwargs.get('x_axis_label','Temperature (K)')
    kwargs['y_axis_label'] = kwargs.get('y_axis_label','Pressure (bars)')
    kwargs['y_axis_type'] = kwargs.get('y_axis_type','log')
    

    ngas = len(out['condensibles'])
    temperature = out['temperature']
    pressure = out['pressure']
    condensibles = out['condensibles']
    mh, mmw = out['scalar_inputs']['mh'], out['scalar_inputs']['mmw']

    kwargs['y_range'] = kwargs.get('y_range',[np.max(pressure), np.min(pressure)])

    fig = figure(**kwargs)

    if with_condensation:
        all_gases = pyeddy.available()
        cond_ts = []
        recommend = []
        line_width = []

        for gas_name in all_gases: #case sensitive names
            #grab p,t from eddysed
            cond_p,t = pyeddy.condensation_t(gas_name, mh, mmw)
            cond_ts +=[t]
            if gas_name in condensibles:
                line_width += [5]
            else: 
                line_width += [1]

        cols = magma(len(all_gases))
        for i in range(len(all_gases)):
            fig.line(cond_ts[i],cond_p, legend_label=all_gases[i],color=cols[i],line_width=line_width[i] )

    fig.line(temperature,pressure, legend_label='User',color='black',line_width=5,line_dash='dashed')

    plot_format(fig)
    if return_condensation: 
        return fig, cond_ts,cond_p,all_gases
    else : 
        return fig


def radii(out,gas=None,at_pressure = 1e-3, compare=False, legend=None, 
        p1w=300, p1h=300, p2w=300, p2h=300, color_indx=0):
    """
    Plots the particle radii profile along with the distribution, at a certain pressure. 

    Parameters
    ----------
    out : dict 
        Dictionary output from pyeddy run 
    pressure : float,optional 
        Pressure level to plot full distribution (bars)
    
    """
    if type(out)==dict:
        out=[out]

    lines = ['solid','dashed','dashdot']
    #lines = ['solid']*3
    if compare:
        lines = ['solid']*len(out)
        legend_it = []
    for j in range(len(out)):
        #compute initial distributions   
        r_g = out[j]['mean_particle_r']
        pressure = out[j]['pressure']

        nl = find_nearest_1d(pressure,at_pressure)

        rmin = out[j]['scalar_inputs']['rmin']
        nrad = out[j]['scalar_inputs']['nrad']
        sig = out[j]['scalar_inputs']['sig']
        ndz = out[j]['column_density']
        #determine which condensed
        which_condensed = [False]*ndz.shape[1]

        # only conisider specified gas
        if gas is not None:
            indx = out[j]['condensibles'].index(gas)
            ndz_gas = np.unique(ndz[:,indx])
            ndz_gas = ndz_gas[ndz_gas>0]
            if len(ndz_gas)>0 : 
                which_condensed[indx] = True
            else:
                print(condensate + " did not condense. Choose another condensate.")
                import sys; sys.exit()

        # consider all gases
        else:
            for i in range(ndz.shape[1]):
                ndz_gas = np.unique(ndz[:,i])
                ndz_gas = ndz_gas[ndz_gas>0]
                if len(ndz_gas)>0 : which_condensed[i] = True


        #take only those that condensed 
        N = ndz[:, which_condensed]
        r_g = r_g[:,which_condensed]
        gas_name = list(np.array(out[j]['condensibles'])[which_condensed])
        r, rup, dr = pyeddy.get_r_grid(r_min = rmin, n_radii = nrad)

        # different colours for different dicts
        if (gas is not None) or compare:
            length = len(out)
        # different colours for different gases
        else:
            length = len(gas_name) 

        color = magma(length)
        color = Colorblind8[color_indx:color_indx+length]

        #initial radii profiles
        df_r_g = {i:r_g[:, gas_name.index(i)] for i in gas_name}
        df_r_g['pressure'] = pressure


        dndr = {}
        for i in gas_name:
            dndr[i]= N[nl,gas_name.index(i)]/r/np.sqrt(2*np.pi)*np.log(sig)*np.exp(
                        - np.log(r/(r_g[nl,gas_name.index(i)]*1e-4))**2/(2*np.log(sig)**2)) #1e-4 is microns to cm
        dndr['r'] = r*1e4 #convert to microns

        if j==0:
            p1 = figure(plot_width=p1w, plot_height=p1h, title="Select Pressure Level", 
                x_axis_type='log',y_axis_type='log',y_axis_label='Pressure (bars)',x_axis_label='Mean Particle Radius (um)',
                y_range=[np.max(pressure),np.min(pressure)])

            p2 = figure(plot_width=p2w, plot_height=p2h, title="Particle Distribution at %.0e bars" %at_pressure,
                         x_axis_type='log', y_axis_type='log', y_axis_label='dn/dr (cm-3)',x_axis_label='Particle Radius (um)')

        #add to r_g for that plot 
        df_r_g['average'] = [pressure[nl]]*len(pressure)
        df_r_g['horizontal'] = np.linspace(np.amin(r_g[r_g>0]), np.amax(r_g) , len(pressure))
            

        s1 = ColumnDataSource(data=dict(df_r_g))
        s2 = ColumnDataSource(data=dict(dndr))
        for i in gas_name:
            if gas is not None or compare: indx = j
            else: indx = gas_name.index(i)
            f = p1.line(i, 'pressure', source=s1, alpha=1,color=color[np.mod(indx, len(color))] ,line_width=4,legend_label=i,line_dash=lines[j])
            p2.line('r', i, source=s2,color=color[np.mod(indx, len(color))] ,line_width=4,line_dash=lines[j])

            if compare:
                legend_it.append((legend[j], [f]))

    # r_g = out['droplet_eff_r']
    # pressure = out['pressure']

    # nl = find_nearest_1d(pressure,at_pressure)

    # rmin = out['scalar_inputs']['rmin']
    # nrad = out['scalar_inputs']['nrad']
    # sig = out['scalar_inputs']['sig']
    # ndz = out['column_density']
    # #determine which condensed
    # which_condensed = [False]*ndz.shape[1]
    # for i in range(ndz.shape[1]):
    #     ndz_gas = np.unique(ndz[:,i])
    #     ndz_gas = ndz_gas[ndz_gas>0]
    #     if len(ndz_gas)>0 : which_condensed[i] = True

    # color = magma(len(which_condensed))

    # #take only those that condensed 
    # N = ndz[:, which_condensed]
    # r_g = r_g[:,which_condensed]
    # gas_name = list(np.array(out['condensibles'])[which_condensed])
    # r, rup, dr = pyeddy.get_r_grid(r_min = rmin, n_radii = nrad)

    # #initial radii profiles
    # df_r_g = {i:r_g[:, gas_name.index(i)] for i in gas_name}
    # df_r_g['pressure'] = pressure


    # #add to r_g for that plot 
    # df_r_g['average'] = [pressure[nl]]*len(pressure)
    # df_r_g['horizontal'] = np.linspace(np.amin(r_g[r_g>0]), np.amax(r_g) , len(pressure))

    p1.line('horizontal', 'average', source=s1, color='black',line_width=3,line_dash='dashed')
    p1.legend.location = 'bottom_left'
    #plot_format(p1)
    #plot_format(p2)
    if compare:
        legend = Legend(items=legend_it, location=(0, 0))
        legend.click_policy="mute"
        p1.add_layout(legend, 'right')   
    return row(p1, p2), dndr

def opd_by_gas(out, gas = None, color = magma, compare=False, legend=None, **kwargs):
    """
    Optical depth for conservative geometric scatteres separated by gas.
    E.g. [Fig 7 in Morley+2012](https://arxiv.org/pdf/1206.4313.pdf)  
    
    Parameters
    ----------
    out : dict 
        Dictionary output from pyeddy run
    color : method
        Method from bokeh.palletes. e.g. bokeh.palletes.virids, bokeh.palletes.magma
    **kwargs : kwargs 
        Kwargs for bokeh.figure() 
    """

    kwargs['plot_height'] = kwargs.get('plot_height',300)
    kwargs['plot_width'] = kwargs.get('plot_width',400)
    kwargs['x_axis_label'] = kwargs.get('x_axis_label','Column Optical Depth')
    kwargs['y_axis_label'] = kwargs.get('y_axis_label','Pressure (bars)')
    kwargs['y_axis_type'] = kwargs.get('y_axis_type','log')
    kwargs['x_axis_type'] = kwargs.get('x_axis_type','log')    

    if type(out)==dict:
        out=[out]

    condensibles = out[0]['condensibles']
    if gas is not None: 
        indx = condensibles.index(gas)
        length = len(out)
    else:
        length = len(condensibles)
    ngas = len(condensibles)
    col = color(length)
    col = Colorblind8[:length]
    lines = ['solid','dashed','dotdash','dashdot']

    legend_it = []
    for j in range(len(out)):
        pressure = out[j]['pressure']
        opd_by_gas = out[j]['opd_by_gas']

        if j == 0:
            kwargs['y_range'] = kwargs.get('y_range',[np.max(pressure), np.min(pressure)])
            kwargs['x_range'] = kwargs.get('x_range',[np.max([1e-6, np.min(opd_by_gas*0.9)]), 
                                                np.max(opd_by_gas*1.1)])
            fig = figure(**kwargs)

        if compare:
            if gas is not None:
                f = fig.line(opd_by_gas[:,indx], pressure, line_width=4, legend_label = condensibles[indx], color=Colorblind8[np.mod(j, len(Colorblind8))])
            else:
                for i in range(ngas):
                    f = fig.line(opd_by_gas[:,i], pressure,line_width=4, legend_label = condensibles[i], color=Colorblind8[np.mod(j, len(Colorblind8))], line_dash=lines[i])
            legend_it.append((legend[j], [f]))
        else:
            if gas is not None:
                fig.line(opd_by_gas[:,indx], pressure, line_width=4, legend_label = condensibles[indx], color=col[j], line_dash=lines[j])
            else:
                for i in range(ngas):
                    fig.line(opd_by_gas[:,i], pressure,line_width=4, legend_label = condensibles[i], color=col[i], line_dash=lines[j])

    if compare:
        legend = Legend(items=legend_it, location=(0, 0))
        legend.click_policy="mute"
        fig.add_layout(legend, 'right')   
    #plot_format(fig)
    return fig

def condensate_mmr(out, gas=None, compare=False, legend=None, **kwargs):
    """
    Condensate mean mass mixing ratio
    
    Parameters
    ----------
    out : dict 
        Dictionary output from pyeddy run
    color : method
        Method from bokeh.palletes. e.g. bokeh.palletes.virids, bokeh.palletes.magma
    **kwargs : kwargs 
        Kwargs for bokeh.figure() 
    """

    kwargs['plot_height'] = kwargs.get('plot_height',300)
    kwargs['plot_width'] = kwargs.get('plot_width',400)
    kwargs['x_axis_label'] = kwargs.get('x_axis_label','Condensate MMR')
    kwargs['y_axis_label'] = kwargs.get('y_axis_label','Pressure (bars)')
    kwargs['y_axis_type'] = kwargs.get('y_axis_type','log')
    kwargs['x_axis_type'] = kwargs.get('x_axis_type','log')    

    if type(out)==dict:
        out=[out]

    condensibles = out[0]['condensibles']
    if gas is not None :
        indx = condensibles.index(gas)
        length = len(out)
    else:
        length = len(condensibles)
    ngas = len(condensibles)
    col = magma(length)
    lines = ['solid','dashed','dotdash','dashdot']

    legend_it = []
    for j in range(len(out)):
        pressure = out[j]['pressure']
        cond_mmr = out[j]['condensate_mmr']

        if j == 0:
            kwargs['y_range'] = kwargs.get('y_range',[np.max(pressure), np.min(pressure)])
            kwargs['x_range'] = kwargs.get('x_range',[np.max([1e-9, np.min(cond_mmr*0.9)]), 
                                                    np.max(cond_mmr*1.1)])
            fig = figure(**kwargs)

        if compare:
            if gas is not None:
                f = fig.line(cond_mmr[:,i], pressure, line_width=4, legend_label = condensibles[indx], color=Colorblind8[np.mod(j, len(Colorblind8))])
            else:
                for i in range(ngas):
                    label = condensibles[i] + ' ' + legend[j]
                    f = fig.line(cond_mmr[:,i], pressure, line_width=4, legend_label = condensibles[i], color=Colorblind8[np.mod(j, len(Colorblind8))], line_dash=lines[i])
            legend_it.append((legend[j], [f]))

        else:
            if gas is not None:
                fig.line(cond_mmr[:,indx], pressure, line_width=4, legend_label = condensibles[indx], color=Colorblind8[np.mod(j, len(Colorblind8))], line_dash=lines[j])
            else:
                for i in range(ngas):
                    fig.line(cond_mmr[:,i], pressure, line_width=4, legend_label = condensibles[i], color=col[i], line_dash=lines[j])

    if compare:
        legend = Legend(items=legend_it, location=(0, 0))
        legend.click_policy="mute"
        fig.add_layout(legend, 'right')   

    plot_format(fig)
    return fig

def all_optics(out):
    """
    Maps of the wavelength dependent single scattering albedo 
    and cloud opacity and asymmetry parameter as a function of altitude. 

    Parameters
    ----------
    out : dict 
        Dictionary output from pyeddy run 

    Returns
    -------
    Three bokeh plots with the single scattering, optical depth, and assymetry maps
    """
    #get into DataFrame format
    dat01 = pyeddy.picaso_format(out['opd_per_layer'],out['single_scattering'],
                      out['asymmetry'])

    nwno=len(out['wave'])
    nlayer=len(out['pressure'])
    pressure=out['pressure']

    pressure_label = 'Pressure (Bars)'

    wavelength_label = 'Wavelength (um)'
    wavelength = out['wave']

    cols = colfun1(200)
    color_mapper = LinearColorMapper(palette=cols, low=0, high=1)

    #PLOT W0
    scat01 = np.flip(np.reshape(dat01['w0'].values,(nlayer,nwno)),0)
    xr, yr = scat01.shape
    f01a = figure(x_range=[0, yr], y_range=[0,xr],
                           x_axis_label=wavelength_label, y_axis_label=pressure_label,
                           title="Single Scattering Albedo",
                          plot_width=300, plot_height=300)


    f01a.image(image=[scat01],  color_mapper=color_mapper, x=0,y=0,dh=xr,dw =yr )

    color_bar = ColorBar(color_mapper=color_mapper, #ticker=LogTicker(),
                       label_standoff=12, border_line_color=None, location=(0,0))

    f01a.add_layout(color_bar, 'left')


    #PLOT OPD
    scat01 = np.flip(np.reshape(dat01['opd'].values,(nlayer,nwno)),0)

    xr, yr = scat01.shape
    cols = colfun2(200)[::-1]
    color_mapper = LogColorMapper(palette=cols, low=1e-3, high=10)


    f01 = figure(x_range=[0, yr], y_range=[0,xr],
                           x_axis_label=wavelength_label, y_axis_label=pressure_label,
                           title="Cloud Optical Depth Per Layer",
                          plot_width=320, plot_height=300)

    f01.image(image=[scat01],  color_mapper=color_mapper, x=0,y=0,dh=xr,dw =yr )

    color_bar = ColorBar(color_mapper=color_mapper, ticker=LogTicker(),
                       label_standoff=12, border_line_color=None, location=(0,0))
    f01.add_layout(color_bar, 'left')

    #PLOT G0
    scat01 = np.flip(np.reshape(dat01['g0'].values,(nlayer,nwno)),0)

    xr, yr = scat01.shape
    cols = colfun3(200)[::-1]
    color_mapper = LinearColorMapper(palette=cols, low=0, high=1)


    f01b = figure(x_range=[0, yr], y_range=[0,xr],
                           x_axis_label=wavelength_label, y_axis_label=pressure_label,
                           title="Asymmetry Parameter",
                          plot_width=300, plot_height=300)

    f01b.image(image=[scat01],  color_mapper=color_mapper, x=0,y=0,dh=xr,dw =yr )

    color_bar = ColorBar(color_mapper=color_mapper, ticker=BasicTicker(),
                       label_standoff=12, border_line_color=None, location=(0,0))
    f01b.add_layout(color_bar, 'left')

    #CHANGE X AND Y AXIS TO BE PHYSICAL UNITS 
    #indexes for pressure plot 
    if (pressure is not None):
        pressure = ["{:.1E}".format(i) for i in pressure[::-1]] #flip since we are also flipping matrices
        npres = len(pressure)
        ipres = np.array(range(npres))
        #set how many we actually want to put on the figure 
        #hard code ten on each.. 
        ipres = ipres[::int(npres/10)]
        pressure = pressure[::int(npres/10)]
        #create dictionary for tick marks 
        ptick = {int(i):j for i,j in zip(ipres,pressure)}
        for i in [f01a, f01, f01b]:
            i.yaxis.ticker = ipres
            i.yaxis.major_label_overrides = ptick
    if (wavelength is not None):
        wave = ["{:.2F}".format(i) for i in wavelength]
        nwave = len(wave)
        iwave = np.array(range(nwave))
        iwave = iwave[::int(nwave/10)]
        wave = wave[::int(nwave/10)]
        wtick = {int(i):j for i,j in zip(iwave,wave)}
        for i in [f01a, f01, f01b]:
            i.xaxis.ticker = iwave
            i.xaxis.major_label_overrides = wtick       

    return gridplot([[f01a, f01,f01b]])

def all_optics_1d(out, wave_range, return_output = False,legend=None,
    colors = colpals.Colorblind8, **kwargs):
    """
    Plots 1d profiles of optical depth per layer, single scattering, and 
    asymmetry averaged over the user input wave_range. 

    Parameters
    ----------
    out : list or dict 
        Either a list of output dictionaries or a single dictionary output
        from .compute(as_dict=True)
    wave_range : list 
        min and max wavelength in microns 
    return_output : bool 
        Default is just to return a figure but you can also 
        return all the 1d profiles 
    legend : bool 
        Default is none. Legend for each component of out 
    **kwargs : keyword arguments
        Key word arguments will be supplied to each bokeh figure function
    """

    kwargs['plot_height'] = kwargs.get('plot_height',300)
    kwargs['plot_width'] = kwargs.get('plot_width',300)
    kwargs['y_axis_type'] = kwargs.get('y_axis_type','log')

    if not isinstance(out, list):
        out = [out]

    pressure = out[0]['pressure']

    kwargs['y_range'] = kwargs.get('y_range',[max(pressure),min(pressure)])     

    ssa = figure(x_axis_label='Single Scattering Albedo',**kwargs)

    g0 = figure(x_axis_label='Asymmetry',**kwargs)

    opd = figure(x_axis_label='Optical Depth',y_axis_label='Pressure (bars)',
        x_axis_type='log',**kwargs)

    
    for i,results in enumerate(out): 
        inds = np.where((results['wave']>wave_range[0]) & 
            (results['wave']<wave_range[1]))

        opd.line(np.mean(results['opd_per_layer'][:,inds],axis=2)[:,0], 
                 results['pressure'], color=colors[np.mod(i, len(colors))],line_width=3)
        
        g0.line(np.mean(results['asymmetry'][:,inds],axis=2)[:,0], 
                 results['pressure'], color=colors[np.mod(i, len(colors))],line_width=3)
        
        if isinstance(legend, type(None)):
            ssa.line(np.mean(results['single_scattering'][:,inds],axis=2)[:,0], 
                 results['pressure'], color=colors[np.mod(i, len(colors))],line_width=3)
        else:
            ssa.line(np.mean(results['single_scattering'][:,inds],axis=2)[:,0], 
                 results['pressure'], color=colors[np.mod(i, len(colors))],line_width=3,
                 legend_label=legend[i])
            ssa.legend.location='top_left'

    if return_output:   
        return gridplot([[opd,ssa,g0]]), [opd,ssa,g0]
    else:   
        return gridplot([[opd,ssa,g0]])

def find_nearest_1d(array,value):
    #small program to find the nearest neighbor in a matrix
    ar , iar ,ic = np.unique(array,return_index=True,return_counts=True)
    idx = (np.abs(ar-value)).argmin(axis=0)
    if ic[idx]>1: 
        idx = iar[idx] + (ic[idx]-1)
    else: 
        idx = iar[idx]
    return idx

def pressure_fig(**plot_kwargs):
    plot_kwargs['y_range'] = plot_kwargs.get('y_range',[1e2,1e-6])
    plot_kwargs['plot_height'] = plot_kwargs.get('plot_height',400)
    plot_kwargs['plot_width'] = plot_kwargs.get('plot_width',600)
    plot_kwargs['x_axis_label'] = plot_kwargs.get('x_axis_label','Temperature (K)')
    plot_kwargs['y_axis_label'] = plot_kwargs.get('y_axis_label','Pressure (bars)')
    plot_kwargs['y_axis_type'] = plot_kwargs.get('y_axis_type','log')        
    fig = figure(**plot_kwargs)
    return fig

def plot_format(df):
    """Function to reformat plots"""
    df.xaxis.axis_label_text_font='times'
    df.yaxis.axis_label_text_font='times'
    df.xaxis.major_label_text_font_size='14pt'
    df.yaxis.major_label_text_font_size='14pt'
    df.xaxis.axis_label_text_font_size='14pt'
    df.yaxis.axis_label_text_font_size='14pt'
    df.xaxis.major_label_text_font='times'
    df.yaxis.major_label_text_font='times'
    df.xaxis.axis_label_text_font_style = 'bold'
    df.yaxis.axis_label_text_font_style = 'bold'

def plot_fsed(pressure, z, scale_height, alpha, beta, epsilon=1e-2, pres_alpha=None, **kwargs):

    kwargs['plot_height'] = kwargs.get('plot_height',400)
    kwargs['plot_width'] = kwargs.get('plot_width',700)
    kwargs['x_axis_label'] = kwargs.get('x_axis_label','fsed')
    kwargs['y_axis_label'] = kwargs.get('y_axis_label','Pressure (bars)')
    kwargs['x_axis_type'] = kwargs.get('x_axis_type','log')
    kwargs['y_axis_type'] = kwargs.get('y_axis_type','log')

    if type(alpha) is int:
        alpha = [alpha]
    if type(beta) is int:
        beta = [beta]

    if pres_alpha is None:
        zstar = max(z)
    else:
        indx = find_nearest_1d(pressure, pres_alpha)
        zstar = z[indx]
    indx = find_nearest_1d(pressure, 1)
    H = scale_height[indx]

    cols = Colorblind8[:len(alpha)*len(beta)]
    lines = ['solid','dashed','dotted','dotdash','dashdot']
    kwargs['y_range'] = kwargs.get('y_range',[np.max(pressure), np.min(pressure)])
    fig = figure(**kwargs)
    
    for i in range(len(alpha)):
        for j in range(len(beta)):
            lab = "alpha=%g" %alpha[i] + ", beta=%g" %beta[j]
            col = Colorblind8[np.mod(i+len(alpha)*j, 8)]
            line = lines[np.mod(j, 5)]
            fsed = alpha[i] * np.exp((z-zstar)/(6*beta[j]*H)) + epsilon
            fig.line(fsed, pressure, legend_label=lab, color=col, line_width=5, line_dash='solid')#line)

    fig.legend.location = "bottom_right"

    return fig

def fsed_from_output(out,labels,y_axis='pressure',color_indx=0,cld_bounds=False,gas_indx=None,**kwargs):

    if type(out)==dict:
        out=[out]

    kwargs['plot_height'] = kwargs.get('plot_height',400)
    kwargs['plot_width'] = kwargs.get('plot_width',700)
    kwargs['x_axis_label'] = kwargs.get('x_axis_label','fsed')
    kwargs['y_axis_label'] = kwargs.get('y_axis_label','Pressure (bars)')
    kwargs['x_axis_type'] = kwargs.get('x_axis_type','log')
    kwargs['y_axis_type'] = kwargs.get('y_axis_type','log')
    #kwargs['x_range'] = kwargs.get('x_range', [1e-2, 2e1])
    if gas_indx is not None:
        title = 'Condensible = ' + str(out[0]['condensibles'][gas_indx])
        kwargs['title'] = kwargs.get('title', title)


    cols = Colorblind8[color_indx:color_indx+len(out)]
    pressure = out[0]['pressure']
    kwargs['y_range'] = kwargs.get('y_range',[np.max(pressure), np.min(pressure)])
    fig = figure(**kwargs)

    min_id=[]; max_id=[]
    for i in range(len(out)):
        if gas_indx is None: x = out[i]['fsed']
        else: x = out[i]['fsed'][:,gas_indx]

        if y_axis is 'pressure':
            y = out[i]['pressure']
        elif y_axis is 'z':
            y = out[i]['altitude']
        col = Colorblind8[np.mod(i+color_indx, 8)]
        fig.line(x, y, legend_label=labels[i], color=col, line_width=5)
        if cld_bounds and gas_indx is not None:
            low_clds = []; high_clds = []
            ndz = out[i]['column_density'][:,gas_indx]
            nonzero = np.where(ndz>1e-3)[0]
            min_id.append(nonzero[0])
            max_id.append(nonzero[-1])
        
    if cld_bounds and gas_indx is not None:
        xmin = kwargs['x_range'][0]
        xmax = kwargs['x_range'][1]
        x1 = np.linspace(xmin,xmax,10)
        y1 = x1*0+y[min(min_id)]
        y2 = x1*0+y[max(max_id)]
        fig.line(x1, y1, color='black',line_width=5, line_dash='dashed')
        fig.line(x1, y2, color='black',line_width=5, line_dash='dashed')


    if labels is not None:
        fig.legend.location = "bottom_left"
    plot_format(fig)
    return fig



def TiO2(mw_atmos, mh=1, gas_mmr = None):
	"""Defines properties for TiO2 as condensible
	
	Parameters 
	----------
	mw_atmos : float 
		Mean molecular weight of the atmosphere amu
	gas_mmr : float , optional
		Gas mass mixing ratio.
		None points to the default value of : 1.69e-7
	mh : float , optional
		Metallicity, Default is 1=1xSolar
	
	Returns
	-------
	mean molecular weight of gas,
	gas mass mixing ratio 
	density of gas cgs
	"""
	if mh != 1: raise Exception("Alert: No M/H Dependence in TiO2 Routine. Consult your local theorist to determine next steps.")
	if isinstance(gas_mmr, type(None)):
		gas_mmr = 1.69e-7 * mh
	gas_mw = 80
	gas_mmr = gas_mmr * (gas_mw/mw_atmos) 
	rho_p =  4.25
	return gas_mw, gas_mmr, rho_p

def CH4(mw_atmos , mh = 1,gas_mmr = None):
	"""Defines properties for CH4 as condensible
	
	Parameters 
	----------
	mw_atmos : float 
		Mean molecular weight of the atmosphere amu
	gas_mmr : float , optional
		Gas mass mixing ratio.
		None points to the default value of : 4.9e-4
	mh : float , optional
		Metallicity, Default is 1=1xSolar
	
	Returns
	-------
	mean molecular weight of gas,
	gas mass mixing ratio 
	density of gas cgs
	"""
	if mh != 1: raise Exception("Alert: No M/H Dependence in CH4 Routine. Consult your local theorist to determine next steps.")
	
	if isinstance(gas_mmr, type(None)):
		gas_mmr = 4.9e-4 * mh

	gas_mw = 16.0
	# Lodders(2003)
	gas_mmr = gas_mmr * (gas_mw/mw_atmos) 
	#V.G. Manzhelii and A.M. Tolkachev, Sov. Phys. Solid State 5, 2506 (1964)
	rho_p =  0.49   #solid (T = 213 K)
	return gas_mw, gas_mmr, rho_p

def NH3(mw_atmos, mh = 1, gas_mmr = None):
	"""Defines properties for NH3 as condensible
	
	Parameters 
	----------
	mw_atmos : float 
		Mean molecular weight of the atmosphere amu
	gas_mmr : float , optional
		Gas mass mixing ratio
		None points to the default value of : 1.34e-4
	mh : float , optional
		Metallicity, Default is 1=1xSolar
	
	Returns
	-------
	mean molecular weight of gas,
	gas mass mixing ratio 
	density of gas cgs
	"""
	if mh != 1: raise Exception("Alert: No M/H Dependence in NH3 Routine. Consult your local theorist to determine next steps.")
	if isinstance(gas_mmr, type(None)):
		gas_mmr = 1.34e-4 * mh

	gas_mw = 17.
	#Lodders(2003)
	gas_mmr = gas_mmr * (gas_mw/mw_atmos) *1.0
	#V.G. Manzhelii and A.M. Tolkachev, Sov. Phys. Solid State 5, 2506 (1964)
	rho_p =  0.84  #solid (T = 213 K)
	return gas_mw, gas_mmr, rho_p

def H2O(mw_atmos, mh = 1, gas_mmr = None):
	"""Defines properties for H2O as condensible
	
	Parameters 
	----------
	mw_atmos : float 
		Mean molecular weight of the atmosphere amu
	gas_mmr : float , optional
		Gas mass mixing ratio
		None points to the default value of : 7.54e-4
	mh : float , optional
		Metallicity, Default is 1=1xSolar
	
	Returns
	-------
	mean molecular weight of gas,
	gas mass mixing ratio 
	density of gas cgs
	"""
	if mh != 1: raise Exception("Alert: No M/H Dependence in H2O Routine. Consult your local theorist to determine next steps.")
	if isinstance(gas_mmr, type(None)):
		gas_mmr = 7.54e-4 * mh
	gas_mw = 18.
	#Lodders(2003)
	gas_mmr = gas_mmr * (gas_mw/mw_atmos)  
	#V.G. Manzhelii and A.M. Tolkachev, Sov. Phys. Solid State 5, 2506 (1964)
	rho_p =  0.93   #solid (T = 213 K)
	return gas_mw, gas_mmr, rho_p

def Fe(mw_atmos,  mh = 1,gas_mmr = None):
	"""Defines properties for Fe as condensible
	
	Parameters 
	----------
	mw_atmos : float 
		Mean molecular weight of the atmosphere amu
	gas_mmr : float , optional
		Gas mass mixing ratio
		None points to the default value of : 5.78e-5
	mh : float , optional
		Metallicity, Default is 1=1xSolar
	
	Returns
	-------
	mean molecular weight of gas,
	gas mass mixing ratio 
	density of gas cgs
	"""
	if mh != 1: raise Exception("Alert: No M/H Dependence in Fe Routine. Consult your local theorist to determine next steps.")
	if isinstance(gas_mmr, type(None)):
		gas_mmr =  5.78e-5 * mh
	gas_mw = 55.845
	gas_mmr = gas_mmr * (gas_mw/mw_atmos)
	#Lodders and Fegley (1998)
	rho_p =  7.875	
	return gas_mw, gas_mmr, rho_p

def KCl(mw_atmos, mh = 1, gas_mmr = None):
	"""Defines properties for KCl as condensible
	
	Parameters 
	----------
	mw_atmos : float 
		Mean molecular weight of the atmosphere amu
	gas_mmr : float , optional
		Gas mass mixing ratio
		None points to the default value of :
		1xSolar = 2.2627E-07
		10xSolar = 2.1829E-06
		50xSolar = 8.1164E-06
	mh : float , optional
		Metallicity, Default is 1=1xSolar
	
	Returns
	-------
	Mean molecular weight of gas,
	Gas mass mixing ratio 
	Density of gas cgs
	"""	
	gas_mw = 74.5

	if isinstance(gas_mmr, type(None)):
		if mh ==1: 
			#SOLAR METALLICITY (abunds tables, 900K, 1 bar)
			gas_mmr = 2.2627E-07 * (gas_mw/mw_atmos)
		elif mh == 10:
			#10x SOLAR METALLICITY (abunds tables, 900K, 1 bar)
			gas_mmr = 2.1829E-06 * (gas_mw/mw_atmos)
		elif mh==50:
			#50x SOLAR METALLICITY (abunds tables, 900K, 1 bar)
			gas_mmr = 8.1164E-06 * (gas_mw/mw_atmos)
		else: 
			raise Exception("KCl gas properties can only be computed for 1, 10 and 50x Solar Meallicity")
	else: 
		gas_mmr = gas_mmr * (gas_mw/mw_atmos)
	#source unknown
	rho_p =  1.99
	return gas_mw, gas_mmr, rho_p

def MgSiO3(mw_atmos, mh = 1, gas_mmr = None):
	"""Defines properties for MgSiO3 as condensible
	
	Parameters 
	----------
	mw_atmos : float 
		Mean molecular weight of the atmosphere amu
	gas_mmr : float , optional
		Gas mass mixing ratio
		None points to the default value of : 2.75e-3
	mh : float , optional
		Metallicity, Default is 1=1xSolar
	
	Returns
	-------
	Mean molecular weight of gas,
	Gas mass mixing ratio 
	Density of gas cgs
	"""	
	if mh != 1: raise Exception("Alert: No M/H Dependence in MgSiO3 Routine. Consult your local theorist to determine next steps.")
	if isinstance(gas_mmr, type(None)):
		gas_mmr =  2.75e-3 * mh # Purposefully no MW scaling
	gas_mw = 100.4

	#Lodders and Fegley (1998)
	rho_p =  3.192
	return gas_mw, gas_mmr, rho_p

def Mg2SiO4(mw_atmos, mh = 1,gas_mmr = None,):
	"""Defines properties for Mg2SiO4 as condensible
	
	Parameters 
	----------
	mw_atmos : float 
		Mean molecular weight of the atmosphere amu
	gas_mmr : float , optional
		Gas mass mixing ratio
		None points to the default value of : 3.58125e-05
	mh : float , optional
		Metallicity, Default is 1=1xSolar
	
	Returns
	-------
	Mean molecular weight of gas,
	Gas mass mixing ratio 
	Density of gas cgs
	"""	
	if mh != 1: raise Exception("Alert: No M/H Dependence in Mg2SiO4 Routine. Consult your local theorist to determine next steps.")
	if isinstance(gas_mmr, type(None)):
		gas_mmr =  59.36e-6 * mh#3.58125e-05
	gas_mw = 140.7
	#NEW FORSTERITE (from Lodders et al. table, 1000mbar, 1900K)
	gas_mmr = gas_mmr * (gas_mw/mw_atmos)  
	#Lodders and Fegley (1998)
	rho_p =  3.214	
	return gas_mw, gas_mmr, rho_p

def MnS(mw_atmos, mh=1,gas_mmr = None):	
	"""Defines properties for MnS as condensible
	
	Parameters 
	----------
	mw_atmos : float 
		Mean molecular weight of the atmosphere amu
	gas_mmr : float , optional
		Gas mass mixing ratio
		None points to the default value of : 6.32e-7
	mh : float , optional
		Metallicity, Default is 1=1xSolar
	
	Returns
	-------
	Mean molecular weight of gas,
	Gas mass mixing ratio 
	Density of gas cgs
	"""	
	if isinstance(gas_mmr, type(None)):
		gas_mmr =  6.32e-7 * mh

	gas_mw = 87.00

	gas_mmr = gas_mmr * (gas_mw/mw_atmos) 

	#Lodders and Fegley (2003) (cvm)
	rho_p =  4.0
	return gas_mw, gas_mmr, rho_p

def ZnS(mw_atmos, mh=1, gas_mmr=None):
	"""Defines properties for ZnS as condensible
	
	Parameters 
	----------
	mw_atmos : float 
		Mean molecular weight of the atmosphere amu
	gas_mmr : float , optional
		Gas mass mixing ratio
		None points to the default value of : 8.40e-8
	mh : float , optional
		Metallicity, Default is 1=1xSolar
	
	Returns
	-------
	Mean molecular weight of gas,
	Gas mass mixing ratio 
	Density of gas cgs
	"""	
	if isinstance(gas_mmr, type(None)):
		gas_mmr =  8.40e-8 * mh

	gas_mw = 97.46

	gas_mmr = gas_mmr * (gas_mw/mw_atmos) 

	#Lodders and Fegley (2003) (cvm)
	rho_p =  4.04	
	return gas_mw, gas_mmr, rho_p

def Cr(mw_atmos, mh=1,gas_mmr=None):
	"""Defines properties for Cr as condensible
	
	Parameters 
	----------
	mw_atmos : float 
		Mean molecular weight of the atmosphere amu
	gas_mmr : float , optional
		Gas mass mixing ratio
		None points to the default value of :
		1xSolar = 8.87e-7
		10xSolar = 8.6803E-06
		50xSolar = 4.1308E-05
	mh : float , optional
		Metallicity, Default is 1=1xSolar
	
	Returns
	-------
	Mean molecular weight of gas,
	Gas mass mixing ratio 
	Density of gas cgs
	"""		
	gas_mw = 51.996
	if isinstance(gas_mmr, type(None)):
		if mh==1:
			gas_mmr = 8.87e-7 * (gas_mw/mw_atmos) 
		elif mh==10:
			gas_mmr = 8.6803E-06 * (gas_mw/mw_atmos) 
		elif mh==50:
			gas_mmr = 4.1308E-05 * (gas_mw/mw_atmos) 
		else: 
			raise Exception("Chromium (Cr) gas properties are only available for 1, 10, and 50xSolar")
	else:
		gas_mmr = gas_mmr * (gas_mw/mw_atmos) 	

	#Lodders and Fegley (2003) (cvm)
	rho_p =  7.15
	return gas_mw, gas_mmr, rho_p

	
def Al2O3(mw_atmos, mh = 1, gas_mmr = None):
	"""Defines properties for Al2O3 as condensible
	
	Parameters 
	----------
	mw_atmos : float 
		Mean molecular weight of the atmosphere amu
	gas_mmr : float , optional
		Gas mass mixing ratio
		None points to the default value of : 2.51e-6
	mh : float , optional
		Metallicity, Default is 1=1xSolar
	
	Returns
	-------
	Mean molecular weight of gas,
	Gas mass mixing ratio 
	Density of gas cgs
	"""		
	if mh != 1: raise Exception("Alert: No M/H Dependence in Al2O3 Routine. Consult your local theorist to determine next steps.")
	if isinstance(gas_mmr, type(None)):
		gas_mmr = 2.51e-6

	gas_mw = 101.961
	#NEW FORSTERITE (from Lodders et al. table, 1000mbar, 1900K)
	gas_mmr = gas_mmr * (gas_mw/mw_atmos)  
	#Lodders and Fegley (1998)
	rho_p =  3.987
	return gas_mw, gas_mmr, rho_p

def Na2S(mw_atmos, mh = 1, gas_mmr = None):
	"""Defines properties for Na2S as condensible
	
	Parameters 
	----------
	mw_atmos : float 
		Mean molecular weight of the atmosphere amu
	gas_mmr : float , optional
		Gas mass mixing ratio
		None points to the default value of : 3.97e-6 
	mh : float , optional
		Metallicity, Default is 1=1xSolar
	
	Returns
	-------
	Mean molecular weight of gas,
	Gas mass mixing ratio 
	Density of gas cgs
	"""		
	if mh != 1: raise Exception("Alert: No M/H Dependence in Na2S Routine. Consult your local theorist to determine next steps.")
	if isinstance(gas_mmr, type(None)):
		gas_mmr = 3.97e-6 
	gas_mw = 78.05
	gas_mmr = gas_mmr * (gas_mw/mw_atmos)  
	#Lodders and Fegley (1998)
	rho_p =  1.856
	return gas_mw, gas_mmr, rho_p




#!/usr/bin/env python

# This sample setup.py can be used as a template for any project using d2to1.
# Simply copy this file and, if desired, delete all the comments.  Also remove
# the 'namespace_packages' and 'packages' arguments to setup.py if this project
# does not contain any packages beloning to a namespace package.

# This import statement attempts to import the setup() function from setuptools
# (this replaces the setup() one uses from distutils.core when using plain
# distutils).
#
# If the import fails (the user doesn't have the setuptools package) it then
# uses the ez_setup bootstrap script to install setuptools, then retries the
# import.  This is common practice for packages using setuptools.
try:
    from setuptools import setup, find_packages
except ImportError:
    from ez_setup import use_setuptools
    use_setuptools()
    from setuptools import setup, find_packages

# The standard setup() call.  Notice, however, that most of the arguments
# normally passed to setup() are absent.  They will instead be read from the
# setup.cfg file using d2to1.
#
# In order for this to work it is necessary to specify setup_requires=['d2to1']
# If the user does not have d2to1, this will boostrap it.  Also require
# stsci.distutils to use any of the common setup_hooks included in
# stsci.distutils (see the setup.cfg for more details).
#
# The next line, which defines namespace_packages and packages is only
# necessary if this projet contains a package belonging to the stsci namespace
# package, such as stsci.distutils or stsci.tools.  This is not necessary for
# projects with their own namespace, such as acstools or pyraf.
#
# d2to1=True is required to enable d2to1 and read the remaning project metadata
# from the setup.cfg file.
#
# use_2to3 and zip_safe are common options support by setuptools; these can
# also be placed in the setup.cfg, as will be demonstrated in a future update
# to this sample package.
setup(
    name='virga-exo', 
    version = '0.3.3',
    description = 'exoplanet code for compute cloud structure',
    long_description = 'README.rst',
    author = 'Natasha E. Batalha',
    author_email = 'natasha.e.batalha@gmail.com',
    url = 'https://natashabatalha.github.io/virga',
    license = 'GPL-3.0',
    download_url = 'https://github.com/natashabatalha/virga',
    classifiers = [
                  'Intended Audience :: Science/Research',
                  'License :: OSI Approved :: BSD License', 
                  'Operating System :: OS Independent' , 
                  'Programming Language :: Python',
                  'Programming Language :: Python :: 3',
                  'Topic :: Scientific/Engineering :: Astronomy',
                  'Topic :: Software Development :: Libraries :: Python Modules'
  ],
  packages=find_packages(),include_package_data=True,
  package_data={'': ['reference/*']},
  install_requires=[
          'numpy',
          'bokeh',
          'pandas',
          'joblib',
          'photutils',
          'astropy',
          'sphinx',
          'scipy',
          'PyMieScatt'
          ], 
    zip_safe = False,
)



from bokeh.palettes import viridis,magma
from bokeh.models import ColumnDataSource, Label, LabelSet,CustomJS
from bokeh.layouts import column,row
from bokeh.plotting import figure, show
from bokeh.models import LinearColorMapper, LogTicker,BasicTicker, ColorBar,LogColorMapper,Legend
from bokeh.palettes import magma as colfun1
from bokeh.palettes import viridis as colfun2
from bokeh.palettes import gray as colfun3
from bokeh.palettes import plasma as colfun4
from bokeh.palettes import Colorblind8
 
import astropy.units as u
import numpy as np

from . import justdoit as pyeddy

def plot_pt(out, labels, lines, **kwargs):

    kwargs['plot_height'] = kwargs.get('plot_height',300)
    kwargs['plot_width'] = kwargs.get('plot_width',600)
    kwargs['x_axis_label'] = kwargs.get('x_axis_label','Temperature (K)')
    kwargs['y_axis_label'] = kwargs.get('y_axis_label','Pressure (bars)')
    kwargs['x_axis_type'] = kwargs.get('x_axis_type','log')
    kwargs['y_axis_type'] = kwargs.get('y_axis_type','log')
    

    cols = viridis(len(out))
    ngas = len(out[0]['condensibles'])
    mh, mmw = out[0]['scalar_inputs']['mh'], out[0]['scalar_inputs']['mmw']
    condensibles = out[0]['condensibles']
    pressure = out[0]['pressure']
    kwargs['y_range'] = kwargs.get('y_range',[np.max(pressure), np.min(pressure)])
    fig = figure(**kwargs)
    for i in range(len(out)):
        temperature = out[i]['temperature']
        pressure = out[i]['pressure']

        fig.line(temperature, pressure, legend_label=labels[i], color=cols[i],line_width=5, line_dash=lines[i])

    plot_format(fig)
    return fig

def plot_cumsum(out,labels,lines,**kwargs):

    kwargs['plot_height'] = kwargs.get('plot_height',300)
    kwargs['plot_width'] = kwargs.get('plot_width',600)
    kwargs['y_axis_label'] = kwargs.get('y_axis_label','Pressure (bars)')
    kwargs['x_axis_type'] = kwargs.get('x_axis_type','log')
    kwargs['y_axis_type'] = kwargs.get('y_axis_type','log')

    cols = viridis(len(out))
    pressure = out[0]['pressure']
    kwargs['y_range'] = kwargs.get('y_range',[np.max(pressure), np.min(pressure)])
    fig = figure(**kwargs)
    for i in range(len(out)):  
        x = np.cumsum(out[i]["opd_per_layer"][:,40])
        pressure = out[i]['pressure']

        fig.line(x, pressure, legend_label=labels[i], color=cols[i],line_width=5, line_dash=lines[i])

    fig.legend.location = "bottom_left"
    plot_format(fig)
    return fig

def plot_output(out,attribute,attribute_label,gas,labels,lines,legend_on=True,
                    color_indx=0,**kwargs):

    condensibles = out[0]['condensibles']
    kwargs['plot_height'] = kwargs.get('plot_height',400)
    kwargs['plot_width'] = kwargs.get('plot_width',350)
    kwargs['x_axis_label'] = kwargs.get('x_axis_label',attribute_label)
    kwargs['y_axis_label'] = kwargs.get('y_axis_label','Pressure (bars)')
    kwargs['x_axis_type'] = kwargs.get('x_axis_type','log')
    kwargs['y_axis_type'] = kwargs.get('y_axis_type','log')
    kwargs['x_range'] = kwargs.get('x_range', [1e-2, 1e4])

    cols = viridis(len(out))
    cols = Colorblind8[color_indx:color_indx+len(out)]
    pressure = out[0]['pressure']
    kwargs['y_range'] = kwargs.get('y_range',[np.max(pressure), np.min(pressure)])
    fig = figure(**kwargs)
    for i in range(len(out)):
        indx = out[i]['condensibles'].index(gas)
        x = out[i][attribute][:,indx]
        if attribute is "column_density":
            x = out[i][attribute][:,indx]/out[i]["layer_thickness"]
        pressure = out[i]['pressure']

        fig.line(x, pressure, legend_label=labels[i], color=cols[i],line_width=5, line_dash=lines[i])

    if legend_on:
        fig.legend.location = "bottom_left"
    plot_format(fig)
    return fig

def opd_by_gas(out,color = magma, **kwargs):
    """
    Optical depth for conservative geometric scatteres separated by gas.
    E.g. [Fig 7 in Morley+2012](https://arxiv.org/pdf/1206.4313.pdf)  
    
    Parameters
    ----------
    out : dict 
        Dictionary output from pyeddy run
    color : method
        Method from bokeh.palletes. e.g. bokeh.palletes.virids, bokeh.palletes.magma
    **kwargs : kwargs 
        Kwargs for bokeh.figure() 
    """

    kwargs['plot_height'] = kwargs.get('plot_height',300)
    kwargs['plot_width'] = kwargs.get('plot_width',400)
    kwargs['x_axis_label'] = kwargs.get('x_axis_label','Column Optical Depth')
    kwargs['y_axis_label'] = kwargs.get('y_axis_label','Pressure (bars)')
    kwargs['y_axis_type'] = kwargs.get('y_axis_type','log')
    kwargs['x_axis_type'] = kwargs.get('x_axis_type','log')    

    ngas = len(out['condensibles'])
    temperature = out['temperature']
    pressure = out['pressure']
    condensibles = out['condensibles']
    opd_by_gas = out['opd_by_gas']

    kwargs['y_range'] = kwargs.get('y_range',[np.max(pressure), np.min(pressure)])
    kwargs['x_range'] = kwargs.get('x_range',[np.max([1e-6, np.min(opd_by_gas*0.9)]), 
                                            np.max(opd_by_gas*1.1)])
    fig = figure(**kwargs)
    col = color(ngas)
    for i in range(ngas):
        fig.line(opd_by_gas[:,i], pressure,line_width=4,legend_label = condensibles[i],color=col[i])

    plot_format(fig)
    return fig

def condensate_mmr(out,color = viridis, **kwargs):
    """
    Condensate mean mass mixing ratio
    
    Parameters
    ----------
    out : dict 
        Dictionary output from pyeddy run
    color : method
        Method from bokeh.palletes. e.g. bokeh.palletes.virids, bokeh.palletes.magma
    **kwargs : kwargs 
        Kwargs for bokeh.figure() 
    """

    kwargs['plot_height'] = kwargs.get('plot_height',300)
    kwargs['plot_width'] = kwargs.get('plot_width',400)
    kwargs['x_axis_label'] = kwargs.get('x_axis_label','Condensate MMR')
    kwargs['y_axis_label'] = kwargs.get('y_axis_label','Pressure (bars)')
    kwargs['y_axis_type'] = kwargs.get('y_axis_type','log')
    kwargs['x_axis_type'] = kwargs.get('x_axis_type','log')    

    ngas = len(out[0]['condensibles'])
    temperature = out['temperature']
    pressure = out['pressure']
    condensibles = out['condensibles']
    cond_mmr = out['condensate_mmr']

    kwargs['y_range'] = kwargs.get('y_range',[np.max(pressure), np.min(pressure)])
    kwargs['x_range'] = kwargs.get('x_range',[np.max([1e-9, np.min(cond_mmr*0.9)]), 
                                            np.max(cond_mmr*1.1)])
    fig = figure(**kwargs)
    col = color(ngas)
    for i in range(ngas):
        fig.line(cond_mmr[:,i], pressure,line_width=4,legend_label = condensibles[i],color=col[i])

    plot_format(fig)
    return fig

def all_optics(out):
    """
    Maps of the wavelength dependent single scattering albedo 
    and cloud opacity and asymmetry parameter as a function of altitude. 

    Parameters
    ----------
    out : dict 
        Dictionary output from pyeddy run 

    Returns
    -------
    Three bokeh plots with the single scattering, optical depth, and assymetry maps
    """
    #get into DataFrame format
    dat01 = pyeddy.picaso_format(out['opd_per_layer'],
                      out['asymmetry'],out['single_scattering'])

    nwno=len(out['wave'])
    nlayer=len(out['pressure'])
    pressure=out['pressure']

    pressure_label = 'Pressure (Bars)'

    wavelength_label = 'Wavelength (um)'
    wavelength = out['wave']

    cols = colfun1(200)
    color_mapper = LinearColorMapper(palette=cols, low=0, high=1)

    #PLOT W0
    scat01 = np.flip(np.reshape(dat01['w0'].values,(nlayer,nwno)),0)
    xr, yr = scat01.shape
    f01a = figure(x_range=[0, yr], y_range=[0,xr],
                           x_axis_label=wavelength_label, y_axis_label=pressure_label,
                           title="Single Scattering Albedo",
                          plot_width=300, plot_height=300)


    f01a.image(image=[scat01],  color_mapper=color_mapper, x=0,y=0,dh=xr,dw =yr )

    color_bar = ColorBar(color_mapper=color_mapper, #ticker=LogTicker(),
                       label_standoff=12, border_line_color=None, location=(0,0))

    f01a.add_layout(color_bar, 'left')


    #PLOT OPD
    scat01 = np.flip(np.reshape(dat01['opd'].values,(nlayer,nwno)),0)

    xr, yr = scat01.shape
    cols = colfun2(200)[::-1]
    color_mapper = LogColorMapper(palette=cols, low=1e-3, high=10)


    f01 = figure(x_range=[0, yr], y_range=[0,xr],
                           x_axis_label=wavelength_label, y_axis_label=pressure_label,
                           title="Cloud Optical Depth Per Layer",
                          plot_width=320, plot_height=300)

    f01.image(image=[scat01],  color_mapper=color_mapper, x=0,y=0,dh=xr,dw =yr )

    color_bar = ColorBar(color_mapper=color_mapper, ticker=LogTicker(),
                       label_standoff=12, border_line_color=None, location=(0,0))
    f01.add_layout(color_bar, 'left')

    #PLOT G0
    scat01 = np.flip(np.reshape(dat01['g0'].values,(nlayer,nwno)),0)

    xr, yr = scat01.shape
    cols = colfun3(200)[::-1]
    color_mapper = LinearColorMapper(palette=cols, low=0, high=1)


    f01b = figure(x_range=[0, yr], y_range=[0,xr],
                           x_axis_label=wavelength_label, y_axis_label=pressure_label,
                           title="Assymetry Parameter",
                          plot_width=300, plot_height=300)

    f01b.image(image=[scat01],  color_mapper=color_mapper, x=0,y=0,dh=xr,dw =yr )

    color_bar = ColorBar(color_mapper=color_mapper, ticker=BasicTicker(),
                       label_standoff=12, border_line_color=None, location=(0,0))
    f01b.add_layout(color_bar, 'left')

    #CHANGE X AND Y AXIS TO BE PHYSICAL UNITS 
    #indexes for pressure plot 
    if (pressure is not None):
        pressure = ["{:.1E}".format(i) for i in pressure[::-1]] #flip since we are also flipping matrices
        npres = len(pressure)
        ipres = np.array(range(npres))
        #set how many we actually want to put on the figure 
        #hard code ten on each.. 
        ipres = ipres[::int(npres/10)]
        pressure = pressure[::int(npres/10)]
        #create dictionary for tick marks 
        ptick = {int(i):j for i,j in zip(ipres,pressure)}
        for i in [f01a, f01, f01b]:
            i.yaxis.ticker = ipres
            i.yaxis.major_label_overrides = ptick
    if (wavelength is not None):
        wave = ["{:.2F}".format(i) for i in wavelength]
        nwave = len(wave)
        iwave = np.array(range(nwave))
        iwave = iwave[::int(nwave/10)]
        wave = wave[::int(nwave/10)]
        wtick = {int(i):j for i,j in zip(iwave,wave)}
        for i in [f01a, f01, f01b]:
            i.xaxis.ticker = iwave
            i.xaxis.major_label_overrides = wtick       

    return row(f01a, f01,f01b)
def find_nearest_1d(array,value):
    #small program to find the nearest neighbor in a matrix
    ar , iar ,ic = np.unique(array,return_index=True,return_counts=True)
    idx = (np.abs(ar-value)).argmin(axis=0)
    if ic[idx]>1: 
        idx = iar[idx] + (ic[idx]-1)
    else: 
        idx = iar[idx]
    return idx
def plot_format(df):
    """Function to reformat plots"""
    df.xaxis.axis_label_text_font='times'
    df.yaxis.axis_label_text_font='times'
    df.xaxis.major_label_text_font_size='14pt'
    df.yaxis.major_label_text_font_size='14pt'
    df.xaxis.axis_label_text_font_size='14pt'
    df.yaxis.axis_label_text_font_size='14pt'
    df.xaxis.major_label_text_font='times'
    df.yaxis.major_label_text_font='times'
    df.xaxis.axis_label_text_font_style = 'bold'
    df.yaxis.axis_label_text_font_style = 'bold'


def plot_fsed(out,labels,y_axis='pressure',color_indx=0,cld_bounds=False,gas_indx=None,**kwargs):

    kwargs['plot_height'] = kwargs.get('plot_height',400)
    kwargs['plot_width'] = kwargs.get('plot_width',700)
    kwargs['x_axis_label'] = kwargs.get('x_axis_label','fsed')
    kwargs['y_axis_label'] = kwargs.get('y_axis_label','Pressure (bars)')
    kwargs['x_axis_type'] = kwargs.get('x_axis_type','log')
    kwargs['y_axis_type'] = kwargs.get('y_axis_type','log')
    #kwargs['x_range'] = kwargs.get('x_range', [1e-2, 2e1])
    if gas_indx is not None:
        title = 'Condensible = ' + str(out[0]['condensibles'][gas_indx])
        kwargs['title'] = kwargs.get('title', title)


    cols = Colorblind8[color_indx:color_indx+len(out)]
    pressure = out[0]['pressure']
    kwargs['y_range'] = kwargs.get('y_range',[np.max(pressure), np.min(pressure)])
    fig = figure(**kwargs)

    min_id=[]; max_id=[]
    for i in range(len(out)):
        if gas_indx is None: x = out[i]['fsed']
        else: x = out[i]['fsed'][:,gas_indx]

        if y_axis is 'pressure':
            y = out[i]['pressure']
        elif y_axis is 'z':
            y = out[i]['altitude']
        col = Colorblind8[np.mod(i+color_indx, 8)]
        fig.line(x, y, legend_label=labels[i], color=col, line_width=5)
        if cld_bounds and gas_indx is not None:
            low_clds = []; high_clds = []
            ndz = out[i]['column_density'][:,gas_indx]
            nonzero = np.where(ndz>1e-3)[0]
            min_id.append(nonzero[0])
            max_id.append(nonzero[-1])
        
    if cld_bounds and gas_indx is not None:
        xmin = kwargs['x_range'][0]
        xmax = kwargs['x_range'][1]
        x1 = np.linspace(xmin,xmax,10)
        y1 = x1*0+y[min(min_id)]
        y2 = x1*0+y[max(max_id)]
        fig.line(x1, y1, color='black',line_width=5, line_dash='dashed')
        fig.line(x1, y2, color='black',line_width=5, line_dash='dashed')


    fig.legend.location = "bottom_left"
    plot_format(fig)
    return fig




import numpy as np
pi=np.pi
import PyMieScatt as ps
import os 

def fort_mie_calc(RO, RFR, RFI, THET, JX, R, RE2, TMAG2, WVNO ):
    """
    Given the refractive indices at a certain wavelength this module
    calculates the Mie scattering by a stratified sphere.The basic code used 
    was that described in the report: " Subroutines for computing the parameters of 
    the electromagnetic radiation scattered by a sphere " J.V. Dave,
    I B M Scientific Center, Palo Alto , California.
    Report NO. 320 - 3236 .. MAY 1968 .

    Parameters
    ----------
    RO : float
        Outer Shell Radius (cm)
    RFR : float
        Real refractive index of shell layer (in the form n= RFR-i*RFI)
    RFI : float
        Imaginary refractive index of shell layer (in the form n= RFR-i*RFI)
    THET : ndarray 
        Angle in degrees between the directions of the incident and the scattered radiation.
    JX : integer
        Total number of THET for which calculations are required
    R : float
        Radius of core (cm)`
    RE2 : float 
        Real refractive index of core (in the form n= RE2-i*TMAG2)
    TMAG2 : float
        Imaginary refractive index of core (in the form n= RE2-i*TMAG2)
        
    WVNO : float
        Wave-number corresponding to the wavelength. (cm^-1)
    
    Returns
    -------
    QEXT: float
        Efficiency factor for extinction,VAN DE HULST,P.14 ' 127
    QSCAT: float
        Efficiency factor for scattering,VAN DE HULST,P.14 ' 127
    CTBQRS: float
        Average(cos(theta))*QSCAT,VAN DE HULST,P.14 ' 127
    ISTATUS: integer
        Convergence indicator, 0 if converged, -1 if otherwise.
        
        
    """
    
    EPSILON_MIE = 1e-7  ## Tolerance for convergence

    nacap, IT = 1000000, 1
    
    ACAP=np.zeros(shape=(nacap),dtype=complex) 
    W=np.zeros(shape=(3,nacap),dtype=complex)
    WFN=np.zeros(shape=(2),dtype=complex)
    Z=np.zeros(shape=(4),dtype=complex)
    U=np.zeros(shape=(8),dtype=complex)
    T= np.zeros(shape=(5))    
    TA=np.zeros(shape=(4))     
    TB=np.zeros(shape=(2))        
    TC=np.zeros(shape=(2))
    TD=np.zeros(shape=(2))  
    TE=np.zeros(shape=(2))  
    PI=np.zeros(shape=(3,IT))   
    TAU=np.zeros(shape=(3,IT))
    CSTHT=np.zeros(shape=(IT)) 
    THETD=np.full((IT),THET) 
    SI2THT=np.zeros(shape=(IT))   
    ELTRMX=np.zeros(shape=(4,IT,2))
    IFLAG = 1
    if  R/RO < 1e-6:    
        IFLAG = 2
    if  JX > IT:
        raise Exception('THE VALUE OF THE ARGUMENT JX IS GREATER THAN IT. PLEASE READ COMMENTS.')
        
            
    RF =  complex(RFR,-RFI)
    RC =  complex( RE2,-TMAG2 )
    X  =  RO * WVNO
    K1 =  RC * WVNO
    K2 =  RF * WVNO
    K3 =  complex( WVNO, 0.0 )
    Z[0] =  K2 * RO
    Z[1] =  K3 * RO
    Z[2] =  K1 * R
    Z[3] =  K2 * R
    X1   =  Z[0].real 
    X4   =  Z[3].real
    Y1   =   Z[0].imag
    Y4   = Z[3].imag
    RRF  =  1.0 / RF
    RX   =  1.0 / X
    RRFX =  RRF * RX
    T[0] = np.sqrt( ( X**2 ) * ( RFR**2 + RFI**2 ))
    NMX1 = int( 1.10 * T[0])
    if  NMX1 > nacap-1 :
        istatus=-1
        return 0,0,0,istatus
        
    NMX2 = int(T[0])
    if  NMX1 <=  150 :
        NMX1 = 150
        NMX2 = 135
    
    ACAP[NMX1]  =  complex( 0.0,0.0 ) #+1
    if IFLAG != 2:
        for N in range(3):
            W[N,NMX1]  = complex( 0.0,0.0 ) #+1
    for N in range(NMX1):
        NN = NMX1 - N-1 ## removed a plus 1 to make up for python loop
        ACAP[NN] = (NN+2) * RRFX - 1.0 / ( (NN+2) * RRFX + ACAP[NN+1] )
        if  IFLAG != 2 :
            for M in range(3):
                W[ M,NN ] = ((NN+2) / Z[M+1])  -1.0 / (  (NN+2) / Z[M+1]  +  W[ M,NN+1 ]  )
    for J in range(JX):
        if  THETD[J] < 0.0 :
            THETD[J] =  abs( THETD[J] )
    
        if  THETD[J] < 90.0:
            T[0]     =  ( 3.14159265359 * THETD[J] ) / 180.0
            CSTHT[J] =  np.cos( T[0])
            SI2THT[J] =  1.0 - CSTHT[J]**2
            continue
        if  THETD[J] == 90.0:
            CSTHT[J]  =  0.0
            SI2THT[J] =  1.0
            continue
        if  THETD[J] > 90.0 :
            raise Exception(' THE VALUE OF THE SCATTERING ANGLE IS GREATER THAN 90.0 DEGREES. PLEASE READ COMMENTS')
        
    for J in range(JX):
        PI[0,J]  =  0.0
        PI[1,J]  =  1.0
        TAU[0,J] =  0.0
        TAU[1,J] =  CSTHT[J]
    T[0]   =  np.cos(X)
    T[1]   =  np.sin(X)
    WM1    =  complex( T[0],-T[1] )
    WFN[0] =  complex( T[1], T[0] )
    TA[0]  =  T[1]
    TA[1]  =  T[0]
    WFN[1] =  RX * WFN[0] - WM1
    TA[2]  =  (WFN[1].real)
    TA[3]  =  (WFN[1].imag)
    if IFLAG != 2:
        N=1
        SINX1   =  np.sin( X1 )
        SINX4   =  np.sin( X4 )
        COSX1   =  np.cos( X1 )
        COSX4   =  np.cos( X4 )
        EY1     =  np.exp( Y1 )
        E2Y1    =  EY1 * EY1
        EY4     =  np.exp( Y4 )
        EY1MY4  =  np.exp( Y1 - Y4 )
        EY1PY4  =  EY1 * EY4
        EY1MY4  =  np.exp( Y1 - Y4 )
        AA  =  SINX4 * ( EY1PY4 + EY1MY4 )
        BB  =  COSX4 * ( EY1PY4 - EY1MY4 )
        CC  =  SINX1 * ( E2Y1 + 1.0 )
        DD  =  COSX1 * ( E2Y1 - 1.0 )
        DENOM   =  1.0  +  E2Y1 * ( 4.0 * SINX1 * SINX1 - 2.0 + E2Y1 )
        REALP   =  ( AA * CC  +  BB * DD ) / DENOM
        AMAGP   =  ( BB * CC  -  AA * DD ) / DENOM
        DUMMY   =  complex( REALP, AMAGP )
        AA  =  SINX4 * SINX4 - 0.5
        BB  =  COSX4 * SINX4
        P24H24  =  0.5 + complex( AA,BB ) * EY4 * EY4
        AA  =  SINX1 * SINX4  -  COSX1 * COSX4
        BB  =  SINX1 * COSX4  +  COSX1 * SINX4
        CC  =  SINX1 * SINX4  +  COSX1 * COSX4
        DD  = -SINX1 * COSX4  +  COSX1 * SINX4
        P24H21  =  0.5 * complex( AA,BB ) * EY1 * EY4 + 0.5 * complex( CC,DD ) * EY1MY4
        DH4  =  Z[3] / ( 1.0 + complex( 0.0,1.0 ) * Z[3] )  -  1.0 / Z[3]
        DH1  =  Z[0] / ( 1.0 + complex( 0.0,1.0 ) * Z[0] )  -  1.0 / Z[0]
        DH2  =  Z[1] / ( 1.0 + complex( 0.0,1.0 ) * Z[1] )  -  1.0 / Z[1]
        PSTORE  =  ( DH4 + N / Z[3] )  *  ( W[2,N-1] + N / Z[3] )
        P24H24  =  P24H24 / PSTORE
        HSTORE  =  ( DH1 + N / Z[0] )  *  ( W[2,N-1] + N / Z[3] )
        P24H21  =  P24H21 / HSTORE
        PSTORE  =  ( ACAP[N-1] + N / Z[0] )  /  ( W[2,N-1] + N / Z[3] )
        DUMMY   =  DUMMY * PSTORE
        DUMSQ   =  DUMMY * DUMMY
        U[0] =  K3 * ACAP[N-1]  -  K2 * W[0,N-1]
        U[1] =  K3 * ACAP[N-1]  -  K2 * DH2
        U[2] =  K2 * ACAP[N-1]  -  K3 * W[0,N-1]
        U[3] =  K2 * ACAP[N-1]  -  K3 * DH2
        U[4] =  K1 *  W[2,N-1]  -  K2 * W[1,N-1]
        U[5] =  K2 *  W[2,N-1]  -  K1 * W[1,N-1]
        U[6] =  complex( 0.0,-1.0 )  *  ( DUMMY * P24H21 - P24H24 )
        U[7] =  TA[2] / WFN[1]
            
        FNA  =  U[7] * ( U[0]*U[4]*U[6]  +  K1*U[0]  -  DUMSQ*K3*U[4] ) /( U[1]*U[4]*U[6]  +  K1*U[1]  -  DUMSQ*K3*U[4] )
        FNB  =  U[7] * ( U[2]*U[5]*U[6]  +  K2*U[2]  -  DUMSQ*K2*U[5] ) /( U[3]*U[5]*U[6]  +  K2*U[3]  -  DUMSQ*K2*U[5] )
        TB[0]=FNA.real
        TB[1]=FNA.imag
        TC[0]=FNB.real
        TC[1]=FNB.imag
    elif IFLAG == 2:
        TC1  =  ACAP[0] * RRF  +  RX
        TC2  =  ACAP[0] * RF   +  RX
        FNA  =  ( TC1 * TA[2]  -  TA[0] ) / ( TC1 * WFN[1]  -  WFN[0] )
        FNB  =  ( TC2 * TA[2]  -  TA[0] ) / ( TC2 * WFN[1]  -  WFN[0] )
        TB[0]=FNA.real
        TB[1]=FNA.imag
        TC[0]=FNB.real
        TC[1]=FNB.imag
        
    FNAP = FNA
    FNBP = FNB
    TD[0]=FNAP.real
    TD[1]=FNAP.imag
    TE[0]=FNBP.real
    TE[1]=FNBP.imag
    T[0] = 1.50
        
    TB[0] = T[0] * TB[0]
    TB[1] = T[0] * TB[1]
    TC[0] = T[0] * TC[0]
    TC[1] = T[0] * TC[1]
        
    for  J in range(JX):
        ELTRMX[0,J,0] = TB[0] * PI[1,J] + TC[0] * TAU[1,J]
        ELTRMX[1,J,0] = TB[1] * PI[1,J] + TC[1] * TAU[1,J]
        ELTRMX[2,J,0] = TC[0] * PI[1,J] + TB[0] * TAU[1,J]
        ELTRMX[3,J,0] = TC[1] * PI[1,J] + TB[1] * TAU[1,J]
        ELTRMX[0,J,1] = TB[0] * PI[1,J] - TC[0] * TAU[1,J]
        ELTRMX[1,J,1] = TB[1] * PI[1,J] - TC[1] * TAU[1,J]
        ELTRMX[2,J,1] = TC[0] * PI[1,J] - TB[0] * TAU[1,J]
        ELTRMX[3,J,1] = TC[1] * PI[1,J] - TB[1] * TAU[1,J]


    QEXT   = 2.0 * ( TB[0] + TC[0])
    QSCAT  = ( TB[0]**2 + TB[1]**2 + TC[0]**2 + TC[1]**2 ) / 0.75
    CTBRQS = 0.0
    QBSR   = -2.0*(TC[0] - TB[0])
    QBSI   = -2.0*(TC[1] - TB[1])
    RMM    = -1.0
    N = 2
    while N <= NMX2:
        T[0] = 2*N - 1
        T[1] =   N - 1
        T[2] = 2*N + 1

        for  J in range(JX):
            PI[2,J]  = ( T[0] * PI[1,J] * CSTHT[J] - N * PI[0,J] ) / T[1]
            TAU[2,J] = CSTHT[J] * ( PI[2,J] - PI[0,J] )  - T[0] * SI2THT[J] * PI[1,J]  +  TAU[0,J]
        WM1    =  WFN[0]
        WFN[0] =  WFN[1]
        TA[0]  =  WFN[0].real

        TA[1]  =  WFN[0].imag
        TA[3]  =  WFN[1].imag
        WFN[1] =  T[0] * RX * WFN[0]  -  WM1
        TA[2]  =  WFN[1].real
        
        
        if IFLAG != 2:
            DH2  =  - N / Z[1]  +  1.0 / ( (N / Z[1]) - DH2 )
            DH4  =  - N / Z[3]  +  1.0 / ( (N / Z[3]) - DH4 )
            DH1  =  - N / Z[0]  +  1.0 / ( (N / Z[0]) - DH1 )
            PSTORE  =  ( DH4 + (N / Z[3] ))  *  ( W[2,N-1] + (N / Z[3] ))
            P24H24  =  P24H24 / PSTORE
            HSTORE  =  ( DH1 + (N / Z[0] ))  *  ( W[2,N-1] + (N / Z[3] ))
            P24H21  =  P24H21 / HSTORE
            PSTORE  =  ( ACAP[N-1] + (N / Z[0] ))  /  ( W[2,N-1] + (N / Z[3] ))
            DUMMY   =  DUMMY * PSTORE
            DUMSQ   =  DUMMY * DUMMY
            U[0] =  K3 * ACAP[N-1]  -  K2 * W[0,N-1]
            U[1] =  K3 * ACAP[N-1]  -  K2 * DH2
            U[2] =  K2 * ACAP[N-1]  -  K3 * W[0,N-1]
            U[3] =  K2 * ACAP[N-1]  -  K3 * DH2
            U[4] =  K1 *  W[2,N-1]  -  K2 * W[1,N-1]
            U[5] =  K2 *  W[2,N-1]  -  K1 * W[1,N-1]
            U[6] =  complex( 0.0,-1.0 )  *  ( DUMMY * P24H21 - P24H24 )
            U[7] =  TA[2] / WFN[1]

            FNA  =  U[7] * ( U[0]*U[4]*U[6]  +  K1*U[0]  -  DUMSQ*K3*U[4] ) /( U[1]*U[4]*U[6]  +  K1*U[1]  -  DUMSQ*K3*U[4] )
            FNB  =  U[7] * ( U[2]*U[5]*U[6]  +  K2*U[2]  -  DUMSQ*K2*U[5] ) /( U[3]*U[5]*U[6]  +  K2*U[3]  -  DUMSQ*K2*U[5] )
            TB[0]=FNA.real
            TB[1]=FNA.imag
            TC[0]=FNB.real
            TC[1]=FNB.imag
        TC1  =  ACAP[N-1] * RRF  +  N * RX
        TC2  =  ACAP[N-1] * RF   +  N * RX
        FN1  =  ( TC1 * TA[2]  -  TA[0] ) /  ( TC1 * WFN[1] - WFN[0] )
        FN2  =  ( TC2 * TA[2]  -  TA[0] ) /  ( TC2 * WFN[1] - WFN[0] )
        M    =  int(WVNO * R)
        if  N >= M : 
            if IFLAG ==2: 
                FNA  =  FN1
                FNB  =  FN2
                TB[0]=FNA.real
                TB[1]=FNA.imag
                TC[0]=FNB.real
                TC[1]=FNB.imag

            if IFLAG != 2:    
                if  abs(  ( FN1-FNA ) / FN1  ) < EPSILON_MIE:
                    if abs(  ( FN2-FNB ) / FN2  )  < EPSILON_MIE :
                        IFLAG = 2
                    
            
     
        T[4]  =  N
        T[3]  =  T[0] / ( T[4] * T[1] )
        T[1]  =  (  T[1] * ( T[4] + 1.0 )  ) / T[4]

        CTBRQS +=  T[1] * ( TD[0] * TB[0]  +  TD[1] * TB[1] + TE[0] * TC[0]  +  TE[1]* TC[1] )+T[3] * ( TD[0] * TE[0]  +  TD[1] * TE[1] )
        QEXT   +=    T[2] * ( TB[0] + TC[0] )
        T[3]    =  TB[0]**2 + TB[1]**2 + TC[0]**2 + TC[1]**2
        QSCAT  +=  T[2] * T[3]
        RMM     =  -RMM
        QBSR +=  T[2]*RMM*(TC[0] - TB[0])
        QBSI  +=  T[2]*RMM*(TC[1] - TB[1])

        T[1]    =  N * (N+1)
        T[0]    =  T[2] / T[1]
        K=int(N)
        for J in range(JX):
            ELTRMX[0,J,0] += T[0]*(TB[0]*PI[2,J]+TC[0]*TAU[2,J])
            ELTRMX[1,J,0] += T[0]*(TB[1]*PI[2,J]+TC[1]*TAU[2,J])
            ELTRMX[2,J,0] += T[0]*(TC[0]*PI[2,J]+TB[0]*TAU[2,J])
            ELTRMX[3,J,0] += T[0]*(TC[1]*PI[2,J]+TB[1]*TAU[2,J])
            if  K%2 == 0:
                ELTRMX[0,J,1] += T[0]*(-TB[0]*PI[2,J]+TC[0]*TAU[2,J])
                ELTRMX[1,J,1] += T[0]*(-TB[1]*PI[2,J]+TC[1]*TAU[2,J])
                ELTRMX[2,J,1] += T[0]*(-TC[0]*PI[2,J]+TB[0]*TAU[2,J])
                ELTRMX[3,J,1] += T[0]*(-TC[1]*PI[2,J]+TB[1]*TAU[2,J])
            else:
                ELTRMX[0,J,1] += T[0]*(TB[0]*PI[2,J]-TC[0]*TAU[2,J])
                ELTRMX[1,J,1] += T[0]*(TB[1]*PI[2,J]-TC[1]*TAU[2,J])
                ELTRMX[2,J,1] += T[0]*(TC[0]*PI[2,J]-TB[0]*TAU[2,J])
                ELTRMX[3,J,1] += T[0]*(TC[1]*PI[2,J]-TB[1]*TAU[2,J])
        
        if  T[3] >= EPSILON_MIE:
            
            N += 1
            for  J in range(JX):
                PI[0,J]   =   PI[1,J]
                PI[1,J]   =   PI[2,J]
                TAU[0,J]  =  TAU[1,J]
                TAU[1,J]  =  TAU[2,J]
             
   
            FNAP  =  FNA
            FNBP  =  FNB
            TD[0]=FNAP.real
            TD[1]=FNAP.imag
            TE[0]=FNBP.real
            TE[1]=FNBP.imag
        
        else:
            break 
    if  N >= NMX2 :
        #print(T[3],NMX2)
        istatus=-1
        return 0,0,0,istatus   
    for J in range(JX):
        for K in range(2):
            for I in range(4):
                T[I]  =  ELTRMX[I,J,K]
    
            ELTRMX[1,J,K]  =      T[0]**2  +  T[1]**2
            ELTRMX[0,J,K]  =      T[2]**2  +  T[3]**2
            ELTRMX[2,J,K]  =  T[0] * T[2]  +  T[1] * T[3]
            ELTRMX[3,J,K]  =  T[1] * T[2]  -  T[3] * T[0]
  
    T[0]    =    2.0 * RX**2
    QEXT    =   QEXT * T[0]
    QSCAT   =  QSCAT * T[0]
    CTBRQS  =  2.0 * CTBRQS * T[0]           
    istatus = 0
    
    return QEXT,QSCAT,CTBRQS,istatus


def calc_new_mieff(wave_in, nn,kk, radius, rup, fort_calc_mie = False):
    ## Calculates optics by reading refrind files
    thetd=0.0   # incident wave angle
    n_thetd=1
    #number of radii sub bins in order to smooth out fringe effects 
    sub_radii = 6
    
    nradii = len(radius)
    nwave=len(wave_in)  #number of wavalength bin centres for calculation
    
    qext = np.zeros((nwave,nradii))
    qscat = np.zeros((nwave,nradii))
    cos_qscat = np.zeros((nwave,nradii))
        
    #compute individual parameters for each gas
    for iwave in range(nwave):
        for irad in range(nradii):
            if irad== 0 :
                dr5= (( rup[0] - radius[0] ) / 5.)
                rr= radius[0]
            else:
                dr5 = ( rup[irad] - rup[irad-1] ) / 5.
                rr  = rup[irad-1]
            corerad = 0.
            corereal = 1.
            coreimag = 0.
        ## averaging over 6 radial bins to avoid fluctuations
            if fort_calc_mie:
                wave=wave_in*1e-4  ## converting to cm 
                bad_mie= False
                for isub in range(sub_radii):
                            wvno=2*np.pi/wave[iwave]
                            qe_pass, qs_pass, c_qs_pass,istatus= fort_mie_calc(rr, nn[iwave], kk[iwave], thetd, n_thetd, corerad, corereal, coreimag, wvno)
                            if istatus == 0:
                                    qe=qe_pass
                                    qs=qs_pass
                                    c_qs=c_qs_pass
                            else:
                                if bad_mie == False :
                                    bad_mie = True
                                    print ('do_optics(): no Mie solution. So previous grid value assigned')
                                    ## The mie_calc routine fails to converge if the real refractive index is smaller than 1. This is true the
                                    ## fortran counterpart as well. So previous step values are assigned

                            qext[iwave,irad]+= qe
                            qscat[iwave,irad]+= qs
                            cos_qscat[iwave,irad] += c_qs
                            rr+=dr5
            #this is the default.
            #if no fortran crappy code, use PyMieScatt which does a much faster 
            #more robust computation of the Mie parameters
            else:
                wave=wave_in*1e3  ## converting to nm 
                ## averaging over 6 radial bins to avoid fluctuations
                for isub in range(sub_radii):
                    #arr = qext, qsca, qabs, g, qpr, qback, qratio
                    arr= ps.MieQCoreShell( corereal+(1j)*coreimag, 
                                            nn[iwave]+(1j)*kk[iwave], 
                                            wave[iwave],dCore=0,dShell=2.0*rr*1e7)

                    qext[iwave,irad]+= arr[0]
                    qscat[iwave,irad]+= arr[1]
                    cos_qscat[iwave,irad] += arr[3]*arr[1] 
                    rr+=dr5

            ## adding to master arrays
            qext[iwave,irad] = qext[iwave,irad] / sub_radii     
            qscat[iwave,irad] = qscat[iwave,irad] / sub_radii
            cos_qscat[iwave,irad] = cos_qscat[iwave,irad]/ sub_radii

    return qext, qscat, cos_qscat


def get_refrind(igas,directory): 
    """
    Reads reference files with wavelength, and refractory indecies. 
    This function relies on input files being structured as a 4 column file with 
    columns: index, wavelength (micron), nn, kk 

    Parameters
    ----------
    igas : str 
        Gas name 
    directory : str 
        Directory were reference files are located. 
    """
    filename = os.path.join(directory ,igas+".refrind")
    idummy, wave_in, nn, kk = np.loadtxt(open(filename,'rt').readlines(), unpack=True, usecols=[0,1,2,3])#[:-1]
    return wave_in,nn,kk


def get_r_grid(r_min=1e-5, n_radii=40):
    """
    Get spacing of radii to run Mie code

    r_min : float 
        Minimum radius to compute (cm)

    n_radii : int
        Number of radii to compute 
    """
    vrat = 2.2 
    pw = 1. / 3.
    f1 = ( 2.0*vrat / ( 1.0 + vrat) )**pw
    f2 = (( 2.0 / ( 1.0 + vrat ) )**pw) * (vrat**(pw-1.0))

    radius = r_min * vrat**(np.linspace(0,n_radii-1,n_radii)/3.)
    rup = f1*radius
    dr = f2*radius

    return radius, rup, dr



import numpy as np

def init_optics(condensibles, nrad=40, rmin=1e-10, read_mie=True): 
	"""
	Setup up a particle size grid and calculate single-particle scattering
	and absorption efficiencies and other parameters to be used by
	`calc_optics()`

	Parameters
	----------
	do_optics : bool 
		(True/False) Calculate optics (T) or use pre computed files (F)
	read_mie : bool 
		(True/False) Read in Mie coefficients from pre compute files `gas_name.mieff`
	condensibles : list of str
		Name in str of all condensible gases e.g. ['H2O','CH4']
	nrad : int 
		Number of radius grid points 
	rmin : float 
		Minimum number of radius grid (cm) 

	Returns
	-------
	wave : array 
		Wavelength bin centers (cm)
	radius : array
		Radius bin centers (cm)
	dr : array 
		Widths of radius bins (cm)
	qscat : array 
		Scattering efficiency 
	qext : array 
		Extinction efficiency
	cos_qscat : array
		qscat * acerage <cos (scattering angle)> 
	"""
	#equations to compute radius bins for particles 
	#these used to be a matrix with different min radii 

	vrat = 2.0
	pw = 1. / 3.
	f1 = ( 2*vrat / ( 1 + vrat) )**pw
	f2 = ( 2 / ( 1 + vrat ) )**pw *  (vrat**pw - 1) 
	radius = rmin * vrat**(irad/3.)
	rup = f1*radius
	dr = f2*radius

	if read_mie: 
		#Read extinction and scattering coefficients 
		#for each condensing vapor
		wave, qscat, qext, cos_qscat = get_meiff()

	else: 
		#Calculate single-scattering efficiencies etc from refractive indices
		#for each condensing vapor

		#Mie parameters:
		#thetd is angle between incident and scattered radiation
		#n_thetd is number of thetd values to consider

		thetd = 0.0
		n_thetd = 1

		#read in refractive indices
		for gas in condensibles: 
			micron_wave, nn, kk = get_refrind(gas)

			cm_wave = micron_wave*1e-4

			wvno = np.pi*2/cm_wave

			for irad in range(nrad):

				#subdivide radius grid into 6 bin to average
				#out oschilations and call Mie code
				if i == 0 : 
					dr5 = (rup[0] - radius[0])/5
					rr = radius[0]
				else:
					dr5 = (rup[irad] - rup[irad-1])/5
					rr = rup[irad-1]

				corerad = 0.0 
				corereal = 1.0
				coreimag = 0.0

				for isub in range(6):
					#should return something that is wave x radius
					qext, qscat, cos_qscat = mie_calc(rr, nn, kk, thetd,
											n_thetd, corerad, corereal, coreimag, wvno)




def get_refrind(condensible, directory='~/Documents/eddysed/input/optics'):
	"""
	Read old style refrind files 

	Parameters
	----------
	condensible : str 
		Condensible name (e.g. Al2O3)

	Returns
	-------
	micron_wave, nn, kk as ndarrays
	"""
	df = pd.read_csv(os.path.join(directory,condensible+'.refrind'),
		skiprows=2, header=None,delim_whitespace=True,
                names=['i', 'wavelength', 'nn', 'kk'])
	micron_wave=df['wavelength'].values
	nn = df['nn'].values
	kk = df['kk'].values
	return micron_wave, nn, kk 









